@@ -1,5 +1,3 @@
-#!/usr/bin/python
-
 # Copyright (C) 2003-2005 Robey Pointer <robey@lag.net>
 #
 # This file is part of paramiko.
@@ -24,6 +22,7 @@ BufferedFile.
 
 from cStringIO import StringIO
 
+
 _FLAG_READ = 0x1
 _FLAG_WRITE = 0x2
 _FLAG_APPEND = 0x4
@@ -32,6 +31,7 @@ _FLAG_BUFFERED = 0x20
 _FLAG_LINE_BUFFERED = 0x40
 _FLAG_UNIVERSAL_NEWLINE = 0x80
 
+
 class BufferedFile (object):
     """
     Reusable base class to implement python-style file buffering around a
@@ -321,7 +321,7 @@ class BufferedFile (object):
             return
         # even if we're line buffering, if the buffer has grown past the
         # buffer size, force a flush.
-        if len(self._wbuffer.getvalue()) >= self._bufsize:
+        if self._wbuffer.tell() >= self._bufsize:
             self.flush()
         return
 
